import numpy as np
from astropy.table import Table, vstack, join
import mesatoolkit as mesat

import matplotlib.pyplot as plt

import shutil
import os

# Important directories
model_dir = '/panfs/ds09/fuller/nrui/mesa-runs/m_plus_wd_models/'
job_dir = '/home/nrui/m_plus_wd/job_shells/'

# Templates
stage1_template = f'{model_dir}templates/stage1_template/'
shell1_template = f'{model_dir}templates/stage1_jobshell.sh'
stage23_template = f'{model_dir}templates/stage23_template/'
shell23_template = f'{model_dir}templates/stage23_jobshell.sh'
stage45_template = f'{model_dir}templates/stage45_template/'
shell45_template = f'{model_dir}templates/stage45_jobshell.sh'
shell1_restart_template = f'{model_dir}templates/stage1_jobshell_restart.sh'
shell23_restart_template = f'{model_dir}templates/stage23_jobshell_restart.sh'
shell45_restart_template = f'{model_dir}templates/stage45_jobshell_restart.sh'


def make_stage1(MinitRG):
    """
    Make m_plus_wd stage1: a red giant with a profile saved for every 0.05 multiple of the helium core

    MinitRG = mass of initial RG

    Terminate when he_core_mass reaches 0.45, since this is the highest core that I'm probing
    """
    # Round quantities for consistency
    MinitRG = round(MinitRG, 2)

    # Copy entire template file to destination
    new_dir = f'{model_dir}stage1_MinitRG{MinitRG:.2f}/'
    shutil.copytree(stage1_template, new_dir)

    # Parameters
    # files_to_change is a dictionary where first item is the template and second is where to write-out (may coincide)
    # parameters is a dictionary of the name of the placeholder keyword (as key) and value (as entry)
    files_to_change = {f'{new_dir}inlist_project': f'{new_dir}inlist_project',
                       shell1_template: f'{job_dir}stage1_MinitRG{MinitRG}.sh',
                       shell1_restart_template: f'{job_dir}stage1_MinitRG{MinitRG}_restart.sh'}
    parameters = {'@MinitRG@': f'{MinitRG:.2f}'}

    for file_to_change in files_to_change.keys():
        f = open(file_to_change, 'r')
        text = f.read()
        f.close()

        for parameter in parameters.keys():
            text = text.replace(parameter, parameters[parameter])

        f = open(files_to_change[file_to_change], 'w')
        f.write(text)
        f.close()

    # Print confirmation message
    print('Successfully created stage1 with:')
    print(f'MinitRG = {MinitRG}')
    print(f'Located at: {new_dir}')

def make_stage23(MinitRG, Mstripped, MenvRetain):
    """
    Make m_plus_wd stage23: stripping of RG into HeWD or sdB (stage2) and evolution of the remnant (stage3)

    MinitRG = mass of initial RG [previous]
    Mstripped = mass of final stripped remnant
    MenvRetain = hydrogen mass to retain, on top of Mstripped
    
    Defines an additional variable MstrippedTotal, which means the total mass of the remnant
    (true mass of helium core of prerequisite plus MenvRetain)
    """
    # Round quantities, for consistency
    MinitRG = round(MinitRG, 2)
    Mstripped = round(Mstripped, 2)
    MenvRetain = round(MenvRetain, 5)

    # Make sure the previous directory exists, and throw error if not
    prerequisite = f'{model_dir}stage1_MinitRG{MinitRG:.2f}/'
    if not os.path.exists(prerequisite):
        raise ValueError(f'Directory {prerequisite} does not exist.')

    # Copy entire template file to destination
    new_dir = f'{model_dir}stage23_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}/'
    shutil.copytree(stage23_template, new_dir)

    # Locate model file from prerequisite
    history = mesat.read_table(f'{prerequisite}LOGS/history.data')
    profiles = mesat.read_profiles(f'{prerequisite}LOGS/profiles.index')
    history = history[np.in1d(history['model_number'], profiles['model_number'])] # keep only those with profiles
    best = np.where(np.abs(history['he_core_mass'] - Mstripped) == np.min(np.abs(history['he_core_mass'] - Mstripped)))[0][0]
    best_model_number = history['model_number'][best]
    best_profile_number = profiles['profile_number'][profiles['model_number'] == best_model_number][0]
    Mstripped_actual = history['he_core_mass'][best] # actual mass of He core of prerequisite model, without rounding (not including MenvRetain)

    shutil.copy(f'{prerequisite}LOGS/profile{best_profile_number}.mod', f'{new_dir}profile_stage2_start.mod')

    # Decide based on MenvRetain which evolution inlist to use
    if MenvRetain <= 5.e-3:
        inlist3_flag = 'WD'
    else:
        inlist3_flag = 'RG'

    # Parameters
    # files_to_change is a dictionary where first item is the template and second is where to write-out (may coincide)
    # parameters is a dictionary of the name of the placeholder keyword (as key) and value (as entry)
    files_to_change = {f'{new_dir}inlist_project_2': f'{new_dir}inlist_project_2',
                       f'{new_dir}inlist_project_3RG': f'{new_dir}inlist_project_3RG',
                       f'{new_dir}inlist_project_3WD': f'{new_dir}inlist_project_3WD',
                       f'{new_dir}inlist_header_3': f'{new_dir}inlist_header_3',
                       shell23_template: f'{job_dir}stage23_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}.sh',
                       shell23_restart_template: f'{job_dir}stage23_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}_restart.sh'}

    parameters = {'@MinitRG@': f'{MinitRG:.2f}',
                  '@Mstripped@': f'{Mstripped:.2f}',
                  '@MenvRetain@': f'{MenvRetain:.5f}',
                  '@MstrippedTotal@': f'{Mstripped_actual + MenvRetain}',
                  '@Inlist3@': inlist3_flag}

    for file_to_change in files_to_change.keys():
        f = open(file_to_change, 'r')
        text = f.read()
        f.close()

        for parameter in parameters.keys():
            text = text.replace(parameter, parameters[parameter])

        f = open(files_to_change[file_to_change], 'w')
        f.write(text)
        f.close()

    # Print confirmation message
    print('Successfully created stage23 with:')
    print(f'MinitRG = {MinitRG}')
    print(f'Mstripped = {Mstripped}')
    print(f'MenvRetain = {MenvRetain}')
    print(f'MstrippedTotal = {Mstripped_actual + MenvRetain}')
    print(f'Located at: {new_dir}')

def make_stage45(MinitRG, Mstripped, MenvRetain, logLinit, MenvFinal, t_range=(1e5, np.inf)):
    """
    Make m_plus_wd stage45: adding the hydrogen back on, and running remnant.

    MinitRG = mass of initial RG [previous]
    Mstripped = mass of final stripped remnant [previous]
    MenvRetain = hydrogen mass to retain, on top of Mstripped [previous]
    logLinit = final ABSOLUTE luminosity of remnant
    MenvFinal = total final mass of hydrogen envelope

    Other important keywords which will not be written explicitly
    t_range = restrict range of logL to look for in prerequisite, in yr (to avoid transient behavior, etc.)
    """
    # Round quantities, for consistency
    MinitRG = round(MinitRG, 2)
    Mstripped = round(Mstripped, 2)
    MenvRetain = round(MenvRetain, 5)
    logLinit = round(logLinit, 3)
    MenvFinal = round(MenvFinal, 2)

    # Make sure the previous directories exists, and throw error if not
    prerequisite1 = f'{model_dir}stage1_MinitRG{MinitRG:.2f}/'
    prerequisite23 = f'{model_dir}stage23_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}/'
    if not os.path.exists(prerequisite1):
        raise ValueError(f'Directory {prerequisite1} does not exist.')
    if not os.path.exists(prerequisite23):
        raise ValueError(f'Directory {prerequisite23} does not exist.')

    # Copy entire template file to destination
    new_dir = f'{model_dir}stage45_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}_logLinit{logLinit:.3f}_MenvFinal{MenvFinal:.2f}/'
    shutil.copytree(stage45_template, new_dir)

    # Locate RG model file from prerequisite1
    history = mesat.read_table(f'{prerequisite1}LOGS/history.data')
    profiles = mesat.read_profiles(f'{prerequisite1}LOGS/profiles.index')
    history = history[np.in1d(history['model_number'], profiles['model_number'])] # keep only those with profiles
    best = np.where(np.abs(history['he_core_mass'] - Mstripped) == np.min(np.abs(history['he_core_mass'] - Mstripped)))[0][0]
    best_model_number = history['model_number'][best]
    best_profile_number = profiles['profile_number'][profiles['model_number'] == best_model_number][0]
    Mstripped_actual = history['he_core_mass'][best] # actual mass of He core of prerequisite model, without rounding (not including MenvRetain)

    shutil.copy(f'{prerequisite1}LOGS/profile{best_profile_number}.mod', f'{new_dir}profile_stage4a_start.mod')

    # Locate core model file from prerequisite23: find the best-matching profile in the t_range
    history = mesat.read_table(f'{prerequisite23}LOGS_stage3/history.data')
    profiles = mesat.read_profiles(f'{prerequisite23}LOGS_stage3/profiles.index')
    history = history[np.in1d(history['model_number'], profiles['model_number'])] # keep only those with profiles
    history = history[history['star_age'] >= t_range[0]]
    history = history[history['star_age'] <= t_range[1]]

    best = np.where(np.abs(history['log_L'] - logLinit) == np.min(np.abs(history['log_L'] - logLinit)))[0][0]
    best_logL = history['log_L'][best]
    best_starage = history['star_age'][best]
    best_model_number = history['model_number'][best]
    best_profile_number = profiles['profile_number'][profiles['model_number'] == best_model_number][0]

    shutil.copy(f'{prerequisite23}LOGS_stage3/profile{best_profile_number}.data', f'{new_dir}profile_core.data')

    # Parameters
    # files_to_change is a dictionary where first item is the template and second is where to write-out (may coincide)
    # parameters is a dictionary of the name of the placeholder keyword (as key) and value (as entry)
    files_to_change = {f'{new_dir}inlist_project_4a': f'{new_dir}inlist_project_4a',
                       f'{new_dir}inlist_pgstar': f'{new_dir}inlist_pgstar',
                       shell45_template: f'{job_dir}stage45_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}_logLinit{logLinit:.3f}_MenvFinal{MenvFinal:.2f}.sh',
                       shell45_restart_template: f'{job_dir}stage45_MinitRG{MinitRG:.2f}_Mstripped{Mstripped:.2f}_MenvRetain{MenvRetain:.5f}_logLinit{logLinit:.3f}_MenvFinal{MenvFinal:.2f}_restart.sh',
                       f'{new_dir}readme': f'{new_dir}readme'}

    Mfinal = Mstripped + MenvFinal
    parameters = {'@MinitRG@': f'{MinitRG:.2f}',
                  '@Mstripped@': f'{Mstripped:.2f}',
                  '@Mplot_min@': f'{Mstripped-0.03:.2f}',
                  '@Mplot_max@': f'{Mstripped+0.03:.2f}',
                  '@Mfinal@': f'{Mfinal:.2f}',
                  '@MenvRetain@': f'{MenvRetain:.5f}',
                  '@logLinit@': f'{logLinit:.3f}',
                  '@MenvFinal@': f'{MenvFinal:.2f}',
                  '@t_range@': f'{t_range}'}

    for file_to_change in files_to_change.keys():
        f = open(file_to_change, 'r')
        text = f.read()
        f.close()

        for parameter in parameters.keys():
            text = text.replace(parameter, parameters[parameter])

        f = open(files_to_change[file_to_change], 'w')
        f.write(text)
        f.close()

    # Print confirmation message
    print('Successfully created stage45 with:')
    print(f'MinitRG = {MinitRG}')
    print(f'Mstripped = {Mstripped}')
    print(f'MenvRetain = {MenvRetain}')
    print(f'logLinit = {best_logL}')
    print(f'MenvFinal = {MenvFinal}')
    print(f'star_age = {best_starage}')
    print(f'Located at: {new_dir}')
