from astropy.table import Table, Column, vstack, unique
import numpy as np

# settings
test_plots_flag = True

min_log_dq_limit = None #-14 # trims off cells of negligibly small mass
max_X_limit = 1e-4 # removes outer hydrogen-rich layers of He WD before stitching



def read_table(fname, header_start=4, remove_duplicates=True):
    """
    Function which takes in the summary.txt file written by MESA/GYRE and return
    an astropy table.

    ARGUMENTS
    ---------
    fname: str
      filename (e.g., 'summary.txt')

    header_start: int (default: 4)
      header start line, default 4 in current MESA/GYRE

    remove_duplicates: bool (default: True)
      if True, removes duplicate rows. this can happen in MESA runs if runs are restarted from photo

    RETURNS
    -------
    tab: astropy.table.Table
      summary table
    """
    f = open(fname, 'r')
    text = f.read().split('\n')
    f.close()

    colname_row = text[header_start+1]
    data_rows = text[header_start+2:-1]

    # Remove all duplicate spaces
    colname_row = ' '.join(colname_row.split())
    data_rows = [' '.join(data_rows[ii].split()) for ii in range(len(data_rows))]

    colnames = colname_row.split()

    # Build each column, assemble table, save
    col_list = []
    existing_colnames = []

    for ii in range(len(colnames)):
        # Read with extra precision
        coldata = [float(data_rows[jj].split()[ii]) for jj in range(len(data_rows))]

        colname = colnames[ii]
        append_num = 0
        while colname in existing_colnames:
            colname = f'{colnames[ii]}{append_num}'
            append_num += 1

        existing_colnames.append(colname)

        col = Column(coldata, name=colname)
        col_list.append(col)

    tab = Table(col_list)
    if remove_duplicates:
        tab = unique(tab)

    return tab

def read_profiles(fname):
    """
    Reads profiles.index
    """
    header_start = 1

    f = open(fname, 'r')
    text = f.read().split('\n')
    f.close()

    data_rows = text[header_start:-1]

    # Remove all duplicate spaces
    data_rows = [' '.join(data_rows[ii].split()) for ii in range(len(data_rows))]

    colnames = ['model_number', 'priority', 'profile_number']

    # Build each column, assemble table, save
    col_list = []

    for ii in range(len(colnames)):
        coldata = [int(data_rows[jj].split()[ii]) for jj in range(len(data_rows))]

        col = Column(coldata, name=colnames[ii])
        col_list.append(col)

    tab = Table(col_list)

    return tab

def stellar_profile_stitch(RG_profile, core_profile):
    """
    Given RG and evolved core (e.g., WD, sdB) profiles, stitch them into a single profile (replacing the core of the former
    with the latter), and return the text of the model file needed to run the model
    """
    Mtot = RG_profile['mass'][0]
    Mcore = core_profile['mass'][0]

    # Add a column to each profile indicating where it came from
    RG_profile.add_column(Column(np.zeros(len(RG_profile)), name='component'))
    core_profile.add_column(Column(np.ones(len(core_profile)), name='component'))

    # Stack catalogs
    profile = vstack([RG_profile[RG_profile['mass'] > Mcore][:-1], core_profile])

    if max_X_limit is not None:
        # Remove hydrogen-rich atmosphere of He WD, where cells tend to be very low
        to_remove = (profile['h1'] >= max_X_limit) & (profile['component'] == 1)
        profile = profile[~to_remove]

    if min_log_dq_limit is not None:
        # Remove cells whose enclosed mass is less than some threshold of the total mass
        # This to avoid bad dq errors due to numerical precision issues
        q_round = Column(np.round(profile['mass'] / profile['mass'][0], -min_log_dq_limit), name='q_round')
        profile.add_column(q_round)
        profile = unique(profile, keys='q_round', keep='first')
        profile = profile[np.argsort(-profile['mass'])] # resort into expected order

    # Recompute all radii, after converting to cgs, to be consistent with new densities
    # the logR column will now be in units of cm, and should not be converted again
    Msun = 1.9892e33 # in g
    profile['mass'] *= Msun

    # extend mass to zero
    ext_mass = np.append(profile['mass'], 0) # Msun
    ext_mass = np.flip(ext_mass)
    delta_mass = np.diff(ext_mass)
    rho = np.flip(10 ** profile['logRho'])

    pi = 3.1415926535897932384626433832795028841971693993751
    delta_cubed_radius = 3. * (delta_mass / rho) / 4. / pi
    cubed_radius = np.cumsum(delta_cubed_radius)
    
    old_logR = np.copy(profile['logR'])
    profile['logR'] = np.flip(np.log10(np.cbrt(cubed_radius)))

    # Compute model file columns
    ln10 = 2.3025850929940455
    Lsun = 3.8418e33

    lnd = profile['logRho'] * ln10
    lnT = profile['logT'] * ln10
    lnR = profile['logR'] * ln10
    L = 10 ** profile['logL'] * Lsun
    dq = -np.diff(np.flip(ext_mass)) / Mtot / Msun
    conv_vel = profile['conv_vel']

    species = ['h1', 'h2', 'he3', 'he4', 'li7', 'be7', 'b8', 'c12', 'c13', 'n13', 'n14', 'n15', 'o14', 'o15', 'o16', 'o17', 'o18', 'f17', 'f18', 'f19', 'ne18', 'ne19', 'ne20', 'ne22', 'mg22', 'mg24']
    species_cols = []

    for specie in species:
        species_cols.append(profile[specie])

    colnames = ['lnd', 'lnT', 'lnR', 'L', 'dq', 'conv_vel'] + species
    coldata = [lnd, lnT, lnR, L, dq, conv_vel] + species_cols

    # Header
    text = '! note: initial lines of file can contain comments\n'
    text += '!\n'
    text += '            36 -- model for mesa/star, with convection velocities (conv_vel). cgs units. lnd=ln(density), lnT=ln(temperature), lnR=ln(radius), L=luminosity, dq=fraction of total mstar in cell; remaining cols are mass fractions.\n'
    text += '\n'

    text += f'                          M/Msun      {Mtot}\n'
    text += '                    model_number                             1\n'
    text += '                        star_age      0.\n'
    text += '                       initial_z      0.02\n'
    text += f'                        n_shells                            {len(profile)}\n'
    text += "                        net_name   'pp_cno_extras_o18_ne22.net'\n"
    text += f'                         species                              {len(species)}\n'
    text += '                            time      0.\n'
    text += '\n'

    spaces = '                '

    text += spaces + spaces.join(colnames)
    text += '\n'

    for ii in range(len(profile)):
        text += f'    {ii+1}'
        for jj, coldatum in enumerate(coldata):
            text += f'    {coldatum[ii]}'
        text += '\n'

    # Populate tail with reasonable parameters
    previous_n_shells = len(profile) # same
    previous_mass = Mtot * Msun # same, in g
    year = 3.1558149984e7
    timestep = 1e2 * year # 10^2 yr, in s
    dt_next = timestep # same

    text += '\n'
    text += '        previous model\n'
    text += '\n'

    text += f'               previous n_shells                            {previous_n_shells}\n'
    text += f'           previous mass (grams)      {previous_mass}\n'
    text += f'              timestep (seconds)      {timestep}\n'
    text += f'               dt_next (seconds)      {dt_next}\n'
    text += '\n'

    if test_plots_flag:
        import matplotlib.pyplot as plt
        
        # Unit conversion
        Rsun = 6.957e10
        logR = profile['logR'] - np.log10(Rsun)

        plt.close()
        plt.figure(figsize=(6, 12))
        
        plt.subplot(611)
        plt.plot(logR, profile['logRho'])
        plt.ylabel('log10(Rho[cgs])')

        plt.subplot(612)
        plt.plot(logR, profile['logT'])
        plt.ylabel('log10(T[K])')

        plt.subplot(613)
        plt.plot(logR, profile['logL'])
        plt.ylabel('log10(L[Lsun])')

        plt.subplot(614)
        plt.plot(logR, np.log10(dq))
        plt.ylabel('log10(dq)')

        plt.subplot(615)
        plt.plot(logR, profile['h1'])
        plt.ylabel('h1')

        plt.subplot(616)
        plt.plot(logR, profile['h1'])
        plt.xlabel('log10(R[Rsun])')
        plt.ylabel('he4')

        plt.tight_layout()
        plt.savefig('./stage4b_plot.png', bbox_inches='tight')

    return text

if __name__ == '__main__':
    RG_profile = read_table('LOGS_stage4a/profile_stage4b_start.data')
    core_profile = read_table('profile_core.data')

    text = stellar_profile_stitch(RG_profile, core_profile)

    f = open('profile_stage5_start.mod', 'w')
    f.write(text)
    f.close()

    print('==============================')
    print('   Created modified profile')
    print('==============================')
